@@ -1,23 +1,27 @@
 /*
- * Copyright (c) 1997-8,2007 Andrew G Morgan <morgan@kernel.org>
+ * Copyright (c) 1997-8,2007-8 Andrew G Morgan <morgan@kernel.org>
  * Copyright (c) 1997 Andrew Main <zefram@dcs.warwick.ac.uk>
  *
  * This file deals with exchanging internal and textual
  * representations of capability sets.
  */
 
+#define _GNU_SOURCE
+#include <stdio.h>
+
 #define LIBCAP_PLEASE_INCLUDE_ARRAY
 #include "libcap.h"
 
 #include <ctype.h>
-#include <stdio.h>
+#include <limits.h>
+#include <fcntl.h>
+#include <unistd.h>
+
+#define SYS_DIR_ROOT_NAMES "/sys/kernel/capability/names"
+#define SYS_DIR_ROOT_CODES "/sys/kernel/capability/codes"
 
 /* Maximum output text length (16 per cap) */
-#define CAP_TEXT_SIZE    (16*__CAP_BITS)
-
-#define LIBCAP_EFF   01
-#define LIBCAP_INH   02
-#define LIBCAP_PER   04
+#define CAP_TEXT_SIZE    (16*__CAP_MAXBITS)
 
 /*
  * Parse a textual representation of capabilities, returning an internal
@@ -26,18 +30,18 @@
 
 #define raise_cap_mask(flat, c)  (flat)[CAP_TO_INDEX(c)] |= CAP_TO_MASK(c)
 
-static void setbits(cap_t a, const __u32 *b, cap_flag_t set)
+static void setbits(cap_t a, const __u32 *b, cap_flag_t set, unsigned blks)
 {
     int n;
-    for (n = __CAP_BLKS; n--; ) {
+    for (n = blks; n--; ) {
 	a->u[n].flat[set] |= b[n];
     }
 }
 
-static void clrbits(cap_t a, const __u32 *b, cap_flag_t set)
+static void clrbits(cap_t a, const __u32 *b, cap_flag_t set, unsigned blks)
 {
     int n;
-    for (n = __CAP_BLKS; n--; )
+    for (n = blks; n--; )
 	a->u[n].flat[set] &= ~b[n];
 }
 
@@ -52,11 +56,11 @@ static char const *namcmp(char const *str, char const *nam)
     return str;
 }
 
-static void forceall(__u32 *flat, __u32 value)
+static void forceall(__u32 *flat, __u32 value, unsigned blks)
 {
     unsigned n;
 
-    for (n = __CAP_BLKS; n--; flat[n] = value);
+    for (n = blks; n--; flat[n] = value);
 
     return;
 }
@@ -71,19 +75,60 @@ static int lookupname(char const **strp)
     str.constp = *strp;
     if (isdigit(*str.constp)) {
 	unsigned long n = strtoul(str.constp, &str.p, 0);
-	if (n >= __CAP_BITS)
+	if (n >= __CAP_MAXBITS)
 	    return -1;
 	*strp = str.constp;
 	return n;
     } else {
+	int fd;
+	char *tmp;
+	int c, n;
+	unsigned len;
+
+	for (len=0; (c = str.constp[len]); ++len) {
+	    if (!(isalpha(c) || (c == '_'))) {
+		break;
+	    }
+	}
+
+#ifdef GPERF_DOWNCASE
+	const struct __cap_token_s *token_info;
+
+	token_info = __cap_lookup_name(str.constp, len);
+	if (token_info != NULL) {
+	    *strp = str.constp + len;
+	    return token_info->index;
+	}
+#else /* ie., ndef GPERF_DOWNCASE */
 	char const *s;
-	int n;
+
 	for (n = __CAP_BITS; n--; )
 	    if (_cap_names[n] && (s = namcmp(str.constp, _cap_names[n]))) {
 		*strp = s;
 		return n;
 	    }
-	return -1;
+#endif /* def GPERF_DOWNCASE */
+
+	asprintf(&tmp, SYS_DIR_ROOT_NAMES "/%s", str.constp);
+	str.constp += len;
+	tmp[sizeof(SYS_DIR_ROOT_NAMES) + len] = '\0';
+
+	fd = open(tmp, O_RDONLY);
+	free(tmp);
+
+	if (fd >= 0) {
+	    char ref[10];
+
+	    len = read(fd, ref, 9);
+	    close(fd);
+	    if (len > 0) {
+		ref[len] = '\0';
+		n = strtoul(ref, NULL, 10);
+		*strp = str.constp;
+		return n;
+	    }
+	}
+	return -1;   	/* No definition available */
     }
 }
 
@@ -91,6 +136,7 @@ cap_t cap_from_text(const char *str)
 {
     cap_t res;
     int n;
+    unsigned cap_blks;
 
     if (str == NULL) {
 	_cap_debug("bad argument");
@@ -101,6 +147,21 @@ cap_t cap_from_text(const char *str)
     if (!(res = cap_init()))
 	return NULL;
 
+    switch (res->head.version) {
+    case _LINUX_CAPABILITY_VERSION_1:
+	cap_blks = _LINUX_CAPABILITY_U32S_1;
+	break;
+    case _LINUX_CAPABILITY_VERSION_2:
+	cap_blks = _LINUX_CAPABILITY_U32S_2;
+	break;
+    case _LINUX_CAPABILITY_VERSION_3:
+	cap_blks = _LINUX_CAPABILITY_U32S_3;
+	break;
+    default:
+	errno = EINVAL;
+	return NULL;
+    }
+    
     _cap_debug("%s", str);
 
     for (;;) {
@@ -108,7 +169,7 @@ cap_t cap_from_text(const char *str)
 	char op;
 	int flags = 0, listed=0;
 
-	forceall(list, 0);
+	forceall(list, 0, __CAP_BLKS);
 
 	/* skip leading spaces */
 	while (isspace((unsigned char)*str))
@@ -126,7 +187,7 @@ cap_t cap_from_text(const char *str)
 	    for (;;) {
 		if (namcmp(str, "all")) {
 		    str += 3;
-		    forceall(list, ~0);
+		    forceall(list, ~0, cap_blks);
 		} else {
 		    n = lookupname(&str);
 		    if (n == -1)
@@ -142,7 +203,7 @@ cap_t cap_from_text(const char *str)
 	} else if (*str == '+' || *str == '-') {
 	    goto bad;                    /* require a list of capabilities */
 	} else {
-	    forceall(list, ~0);
+	    forceall(list, ~0, cap_blks);
 	}
 
 	/* identify first operation on list of capabilities */
@@ -181,28 +242,28 @@ cap_t cap_from_text(const char *str)
 	    case '=':
 	    case 'P':                                              /* =+ */
 	    case 'M':                                              /* =- */
-		clrbits(res, list, CAP_EFFECTIVE);
-		clrbits(res, list, CAP_PERMITTED);
-		clrbits(res, list, CAP_INHERITABLE);
+		clrbits(res, list, CAP_EFFECTIVE, cap_blks);
+		clrbits(res, list, CAP_PERMITTED, cap_blks);
+		clrbits(res, list, CAP_INHERITABLE, cap_blks);
 		if (op == 'M')
 		    goto minus;
 		/* fall through */
 	    case '+':
 		if (flags & LIBCAP_EFF)
-		    setbits(res, list, CAP_EFFECTIVE);
+		    setbits(res, list, CAP_EFFECTIVE, cap_blks);
 		if (flags & LIBCAP_PER)
-		    setbits(res, list, CAP_PERMITTED);
+		    setbits(res, list, CAP_PERMITTED, cap_blks);
 		if (flags & LIBCAP_INH)
-		    setbits(res, list, CAP_INHERITABLE);
+		    setbits(res, list, CAP_INHERITABLE, cap_blks);
 		break;
 	    case '-':
 	    minus:
 		if (flags & LIBCAP_EFF)
-		    clrbits(res, list, CAP_EFFECTIVE);
+		    clrbits(res, list, CAP_EFFECTIVE, cap_blks);
 		if (flags & LIBCAP_PER)
-		    clrbits(res, list, CAP_PERMITTED);
+		    clrbits(res, list, CAP_PERMITTED, cap_blks);
 		if (flags & LIBCAP_INH)
-		    clrbits(res, list, CAP_INHERITABLE);
+		    clrbits(res, list, CAP_INHERITABLE, cap_blks);
 		break;
 	    }
 
@@ -228,6 +289,56 @@ bad:
     return res;
 }
 
+/*
+ * lookup a capability name and return its numerical value
+ */
+int cap_from_name(const char *name, cap_value_t *value_p)
+{
+    int n;
+
+    if (((n = lookupname(&name)) >= 0) && (value_p != NULL)) {
+	*value_p = (unsigned) n;
+    }
+    return -(n < 0);
+}
+
+/*
+ * Convert a single capability index number into a string representation
+ */
+char *cap_to_name(cap_value_t cap)
+{
+    if ((cap < 0) || (cap >= __CAP_BITS)) {
+#if UINT_MAX != 4294967295U
+# error Recompile with correctly sized numeric array
+#endif
+	char *tmp, *result;
+	int fd, len;
+
+	asprintf(&tmp, SYS_DIR_ROOT_CODES "/%u", cap);
+	fd = open(tmp, O_RDONLY);
+
+	if (fd >= 0) {
+	    len = read(fd, tmp, sizeof(SYS_DIR_ROOT_CODES));
+	    close(fd);
+	    if (len > 0) {
+		tmp[len-1] = '\0';
+	    } else {
+		goto default_to_numeric;
+	    }
+	} else {
+	default_to_numeric:
+	    sprintf(tmp, "%u", cap);
+	}
+
+	result = _libcap_strdup(tmp);
+	free(tmp);
+
+	return result;
+    } else {
+	return _libcap_strdup(_cap_names[cap]);
+    }
+}
+
 /*
  * Convert an internal representation to a textual one. The textual
  * representation is stored in static memory. It will be overwritten
@@ -255,10 +366,12 @@ static int getstateflags(cap_t caps, int capno)
 
 char *cap_to_text(cap_t caps, ssize_t *length_p)
 {
-    static char buf[CAP_TEXT_SIZE+CAP_TEXT_BUFFER_ZONE];
+    char buf[CAP_TEXT_SIZE+CAP_TEXT_BUFFER_ZONE];
     char *p;
-    int histo[8] = {0};
-    int m, n, t;
+    int histo[8];
+    int m, t;
+    unsigned n;
+    unsigned cap_maxbits, cap_blks;
 
     /* Check arguments */
     if (!good_cap_t(caps)) {
@@ -266,17 +379,44 @@ char *cap_to_text(cap_t caps, ssize_t *length_p)
 	return NULL;
     }
 
+    switch (caps->head.version) {
+    case _LINUX_CAPABILITY_VERSION_1:
+	cap_blks = _LINUX_CAPABILITY_U32S_1;
+	break;
+    case _LINUX_CAPABILITY_VERSION_2:
+	cap_blks = _LINUX_CAPABILITY_U32S_2;
+	break;
+    case _LINUX_CAPABILITY_VERSION_3:
+	cap_blks = _LINUX_CAPABILITY_U32S_3;
+	break;
+    default:
+	errno = EINVAL;
+	return NULL;
+    }
+
+    cap_maxbits = 32 * cap_blks;
+
     _cap_debugcap("e = ", *caps, CAP_EFFECTIVE);
     _cap_debugcap("i = ", *caps, CAP_INHERITABLE);
     _cap_debugcap("p = ", *caps, CAP_PERMITTED);
 
-    for (n = __CAP_BITS; n--; )
+    memset(histo, 0, sizeof(histo));
+
+    /* default prevailing state to the upper - unnamed bits */
+    for (n = cap_maxbits-1; n > __CAP_BITS; n--)
 	histo[getstateflags(caps, n)]++;
 
+    /* find which combination of capability sets shares the most bits */
     for (m=t=7; t--; )
 	if (histo[t] > histo[m])
 	    m = t;
 
+    /* capture remaining bits - selecting m from only the unnamed bits,
+       we maximize the likelihood that we won't see numeric capability
+       values in the text output. */
+    while (n--)
+	histo[getstateflags(caps, n)]++;
+
     /* blank is not a valid capability set */
     p = sprintf(buf, "=%s%s%s",
 		(m & LIBCAP_EFF) ? "e" : "",
@@ -286,16 +426,18 @@ char *cap_to_text(cap_t caps, ssize_t *length_p)
     for (t = 8; t--; )
 	if (t != m && histo[t]) {
 	    *p++ = ' ';
-	    for (n = 0; n != __CAP_BITS; n++)
+	    for (n = 0; n < cap_maxbits; n++)
 		if (getstateflags(caps, n) == t) {
-		    if (_cap_names[n])
-			p += sprintf(p, "%s,", _cap_names[n]);
-		    else
-			p += sprintf(p, "%d,", n);
-		    if (p - buf > CAP_TEXT_SIZE) {
+		    char *this_cap_name;
+
+		    this_cap_name = cap_to_name(n);
+		    if ((strlen(this_cap_name) + (p - buf)) > CAP_TEXT_SIZE) {
+			cap_free(this_cap_name);
 			errno = ERANGE;
 			return NULL;
 		    }
+		    p += sprintf(p, "%s,", this_cap_name);
+		    cap_free(this_cap_name);
 		}
 	    p--;
 	    n = t & ~m;
